﻿using System;
using System.Collections.Generic;
using System.Runtime.InteropServices;
using System.Text;
using System.Runtime.Serialization;

namespace Lexalytics
{
    public class SalienceFiveCollection : SalienceFiveEngine, IDisposable
    {
#region MANAGED .NET LAYER

    #region Collection-specific options
        private float m_NeutralSentimentUpperBound;
        /// <summary>
        /// Float defining the upper bound of a sentiment value to be considered neutral, default value is 0.5
        /// </summary>
        public float NeutralSentimentUpperBound
        {
            get { return m_NeutralSentimentUpperBound; }
            set { if (SetSalienceOption(4002, value) == 0) m_NeutralSentimentUpperBound = value; }
        }

        private float m_NeutralSentimentLowerBound;
        /// <summary>
        /// Float defining the lower bound of a sentiment value to be considered neutral, default value is -0.45
        /// </summary>
        public float NeutralSentimentLowerBound
        {
            get { return m_NeutralSentimentLowerBound; }
            set { if (SetSalienceOption(4003, value) == 0) m_NeutralSentimentLowerBound = value; }
        }

        private int m_CollectionResultSize;
        /// <summary>
        /// Defines maximum number of results to return for collection functions such as facets or entities, default value is 20
        /// </summary>
        public int CollectionResultSize
        {
            get { return m_CollectionResultSize; }
            set { if (SetSalienceOption(6000, value) == 0) m_CollectionResultSize = value; }
        }

        private bool m_UseSemantics;
        /// <summary>
        /// Determines whether or not to utilize the concept matrix during collection operations
        /// </summary>
        public bool UseSemantics
        {
            get { return m_UseSemantics; }
            set { if (SetSalienceOption(6001, value) == 0) m_UseSemantics = value; }
        }

    #endregion // Collection-specific options

    #region Constructors

        /// <summary>
        /// Standard constructor, requires path to license file and path to data directory
        /// </summary>
        /// <param name="sLicensePath">Path to a valid Lexalytics license file</param>
        /// <param name="sDataPath">Path to data directory to use for this session</param>
        /// <exception cref="SalienceException">
        /// Thrown if license file cannot be loaded or error in loading data directory (AccessViolationException), or unable to load underlying Salience Engine (DllNotFoundException)
        /// </exception>
        /// <see cref="http://dev.lexalytics.com/wiki-se5/pmwiki.php?n=SE5API.SalienceFiveCollectionConstructor"/>
        public SalienceFiveCollection(string sLicensePath, string sDataPath) : base(sLicensePath, sDataPath) {}

        /// <summary>
        /// Standard constructor, requires path to license file and path to data directory
        /// </summary>
        /// <param name="sLicensePath">Path to a valid Lexalytics license file</param>
        /// <param name="sDataPath">Path to data directory to use for this session</param>
        /// <param name="sUserDir">Path to a user directory of customizations to use for this session</param>
        /// <exception cref="SalienceException">
        /// Thrown if license file cannot be loaded or error in loading data directory (AccessViolationException), or unable to load underlying Salience Engine (DllNotFoundException)
        /// </exception>
        /// <see cref="http://dev.lexalytics.com/wiki-se5/pmwiki.php?n=SE5API.SalienceFiveCollectionConstructor"/>
        public SalienceFiveCollection(string sLicensePath, string sDataPath, string sUserDir) : base(sLicensePath, sDataPath, sUserDir) { }

        /// <summary>
        /// Standard constructor, requires path to license file and path to data directory
        /// </summary>
        /// <param name="sLicensePath">Path to a valid Lexalytics license file</param>
        /// <param name="sDataPath">Path to data directory to use for this session</param>
        /// <param name="sUserDir">Path to a user directory of customizations to use for this session</param>
        /// <param name="sLogPath">Path to a log file to write session in itialization information to</param>
        /// <exception cref="SalienceException">
        /// Thrown if license file cannot be loaded or error in loading data directory (AccessViolationException), or unable to load underlying Salience Engine (DllNotFoundException)
        /// </exception>
        /// <see cref="http://dev.lexalytics.com/wiki-se5/pmwiki.php?n=SE5API.SalienceFiveCollectionConstructor"/>
        public SalienceFiveCollection(string sLicensePath, string sDataPath, string sUserDir, string sLogPath) : base(sLicensePath, sDataPath, sUserDir, sLogPath) { }

        /// <summary>
        /// Standard constructor, requires path to license file and path to data directory
        /// </summary>
        /// <param name="sLicensePath">Path to a valid Lexalytics license file</param>
        /// <param name="oStartup">Startup object giving other session startup parameters</param>
        /// <exception cref="SalienceException">
        /// Thrown if license file cannot be loaded or error in loading data directory (AccessViolationException), or unable to load underlying Salience Engine (DllNotFoundException)
        /// </exception>
        /// <see cref="http://dev.lexalytics.com/wiki-se5/pmwiki.php?n=SE5API.SalienceFiveCollectionConstructor"/>
        public SalienceFiveCollection(string sLicensePath, SalienceStartup oStartup) : base(sLicensePath, oStartup) { }

    #endregion

    #region Object disposal
        new public void Dispose()
        {
            base.Dispose();
        }
    #endregion 

    #region Utility methods
        /// <summary>
        /// Wrapper method to return the Salience Engine version
        /// </summary>
        /// <returns>String with Salience Engine version information</returns>
        /// <remarks>Wrapper around C API method lxaGetSalienceVersion</remarks>
        /// <exception cref="SalienceException">Thrown if version information is not available</exception>
        /// <see cref="http://dev.lexalytics.com/wiki-se5/pmwiki.php?n=SE5API.SalienceFiveCollectionGetSalienceVersion"/>
        new public string GetSalienceVersion()
        {
            return base.GetSalienceVersion();
        }

        /// <summary>
        /// Static method for retrieving the default Salience location without needing to instantiate an object
        /// </summary>
        /// <returns>String path to Salience installation</returns>
        new public static string GetDefaultSalienceLocation()
        {
            return SalienceFiveEngine.GetDefaultSalienceLocation();
        }

        /// <summary>
        /// Wrapper method to return the output of Salience Engine operating environment
        /// </summary>
        /// <returns>String with Salience Engine environment information</returns>
        /// <remarks>Wrapper around C API method lxaDumpEnvironment</remarks>
        /// <exception cref="SalienceException">Thrown if version information is not available</exception>
        /// <see cref="http://dev.lexalytics.com/wiki-se5/pmwiki.php?n=SE5API.SalienceFiveCollectionDumpEnvironment"/>
        new public string DumpEnvironment()
        {
            return base.DumpEnvironment();
        }

    #endregion // Utility methods

    #region Text preparation methods
        /// <summary>
        /// Prepares a collection from a list collection of strings
        /// </summary>
        /// <returns>Integer error return code</returns>
        /// <remarks>Wrapper around C API method lxaPrepareCollection</remarks>
        /// <exception cref="SalienceException">Thrown if text cannot be parsed</exception>
        /// <see cref="http://dev.lexalytics.com/wiki-se5/pmwiki.php?n=SE5API.SalienceFiveCollectionPrepareCollectionFromList"/>
        new public int PrepareCollectionFromList(String sName,List<String> lstText)
        {
            return base.PrepareCollectionFromList(sName, lstText);
        }

        /// <summary>
        /// Prepares a collection from a file containing individual pieces of content, one per line
        /// </summary>
        /// <param name="sFile">Full path to a readable text file</param>
        /// <param name="sName">Name for the collection</param>
        /// <returns>Integer error return code</returns>
        /// <remarks>Wrapper around C API method lxaPrepareCollectionFromFile</remarks>
        /// <exception cref="SalienceException">Thrown if text cannot be parsed</exception>
        /// <see cref="http://dev.lexalytics.com/wiki-se5/pmwiki.php?n=SE5API.SalienceFiveCollectionPrepareCollectionFromFile"/>
        new public int PrepareCollectionFromFile(string sName, string sFile)
        {
            return base.PrepareCollectionFromFile(sName, sFile);
        }

    #endregion 

    #region Details methods
        /// <summary>
        /// Wrapper method to return details about the current collection
        /// </summary>
        /// <returns>SalienceCollectionDetails object</returns>
        /// <remarks>Wrapper around C API method lxaGetCollectionDetails</remarks>
        /// <see cref="http://dev.lexalytics.com/wiki-se5/pmwiki.php?n=SE5API.SalienceFiveCollectionGetCollectionDetails"/>
        new public SalienceCollectionDetails GetCollectionDetails()
        {
            return base.GetCollectionDetails();
        }

    #endregion // Details methods

    #region Collection level functions

        /// <summary>
        /// Returns a list of themes extracted from the collection
        /// </summary>
        /// <returns>A List of SalienceTheme objects, each containing information about a theme extracted from the collection</returns>
        /// <remarks>Wrapper around C API method lxaGetCollectionThemes</remarks>
        /// <see cref="http://dev.lexalytics.com/wiki-se5/pmwiki.php?n=SE5API.SalienceFiveCollectionGetThemes"/>
        public List<SalienceTheme> GetThemes()
        {
            return base.GetCollectionThemes();
        }

        /// <summary>
        /// Returns a list of entities extracted from the collection
        /// </summary>
        /// <returns>A List of SalienceCollectionEntity objects, each containing information about an entity extracted from the collection</returns>
        /// <remarks>Wrapper around C API method lxaGetCollectionEntities</remarks>
        /// <see cref="http://dev.lexalytics.com/wiki-se5/pmwiki.php?n=SE5API.SalienceFiveCollectionGetEntities"/>
        public List<SalienceCollectionEntity> GetEntities()
        {
            return base.GetCollectionEntities();
        }

        /// <summary>
        /// Returns a list of facets extracted from the collection
        /// </summary>
        /// <returns>A List of SalienceFacet objects, each containing information about a facet and its attributed extracted from the collection</returns>
        /// <remarks>Wrapper around C API method lxaGetCollectionFacets</remarks>
        /// <see cref="http://dev.lexalytics.com/wiki-se5/pmwiki.php?n=SE5API.SalienceFiveCollectionGetFacets"/>
        public List<SalienceFacet> GetFacets()
        {
            return base.GetCollectionFacets();
        }

        /// <summary>
        /// Returns a list of topics extracted from the collection via tag query definitions.
        /// </summary>
        /// <returns>A List of SalienceTopic objects, each containing information about a topic identified for at least one document in the collection</returns>
        /// <remarks>Wrapper around C API method lxaGetCollectionQueryDefinedTopics</remarks>
        /// <see cref="http://dev.lexalytics.com/wiki-se5/pmwiki.php?n=SE5API.SalienceFiveCollectionGetQueryDefinedTopics"/>
        new public List<SalienceTopic> GetQueryDefinedTopics()
        {
            return base.GetCollectionQueryDefinedTopics();
        }

        /// <summary>
        /// Returns a list of topics extracted from the collection via the concept matrix.
        /// </summary>
        /// <returns>A List of SalienceTopic objects, each containing information about a topic identified for at least one document in the collection</returns>
        /// <remarks>Wrapper around C API method lxaGetCollectionConceptDefinedTopics</remarks>
        /// <see cref="http://dev.lexalytics.com/wiki-se5/pmwiki.php?n=SE5API.SalienceFiveCollectionGetConceptTopics"/>
        public List<SalienceTopic> GetConceptTopics()
        {
            return base.GetCollectionConceptTopics();
        }

    #endregion // Collection level functions
#endregion // MANAGED .NET LAYER

    }
}
