﻿using System;
using System.Collections.Generic;
using System.Runtime.InteropServices;
using System.Text;
using System.Runtime.Serialization;

namespace Lexalytics
{
    public class SalienceFiveEngine : IDisposable
    {
        private const int LXA_MAX_ERROR_LENGTH = 1024;
        protected bool m_bDisposed = false;
        protected IntPtr m_pSessionHandle;
        protected IntPtr m_pLicense;
        public delegate int dEngineNotification(int nMessageId, string sMessage);
        protected umSalienceCallback fCallback;
        public dEngineNotification Callback_GetCollectionFacets;
        public dEngineNotification Callback_GetCollectionEntities;
        public dEngineNotification Callback_GetCollectionThemes;
        public dEngineNotification Callback_PrepareCollection;

        #region  Base options
        private int m_TextThreshold;
        /// <summary>
        /// Integer property indicating the required threshold (between 0 and 100) of text that must be alphanumeric characters for text processing. 
        /// By default, 80% of the content sumbitted to a text preparation method must be alphanumeric. Otherwise, a "Failed to parse" exception is thrown.
        /// </summary>
        public int TextThreshold
        {
            get { return m_TextThreshold; }
            set { if (SetSalienceOption(1000,value)==0) m_TextThreshold=value; }
        }

        private int m_ExecutionTimeout;
        /// <summary>
        /// Maximum time in milliseconds certain functions are allowed to run for (0=no timeout)
        /// </summary>
        public int ExecutionTimeout
        {
            get { return m_ExecutionTimeout; }
            set { if (SetSalienceOption(1002, value) == 0) m_ExecutionTimeout = value; }
        }

        private bool m_FailOnLongSentences;
        /// <summary>
        /// When a document contains a very long sentence, should you process the rest of the document? (0=process)
        /// </summary>
        public bool FailOnLongSentences
        {
            get { return m_FailOnLongSentences; }
            set { if (SetSalienceOption(1003, value) == 0) m_FailOnLongSentences = value; }
        }

        private string m_UserDirectory;
        /// <summary>
        /// Sets path to user directory of customizations
        /// </summary>
        public string UserDirectory
        {
            get { return m_UserDirectory; }
            set { if (SetSalienceOption(1004, value) == 0) m_UserDirectory = value; }
        }

        private float m_ConceptSlop;
        /// <summary>
        /// Sets a threshold for the measure of similarity between two concepts for them to roll together. This is a float value between 0 and 1, default value is 0.8. We do not recommend significant changes to this option value.
        /// </summary>
        public float ConceptSlop
        {
            get { return m_ConceptSlop; }
            set { if (SetSalienceOption(1005, value) == 0) m_ConceptSlop = value; }
        }

        #endregion // Base options

        #region Concept options
        private int m_MaxConceptHits;
        public int MaxConceptHits
        {
            get { return m_MaxConceptHits; }
            set { if (SetSalienceOption(2000, value) == 0) m_MaxConceptHits = value; }
        }

        private float m_MinConceptScore;
        public float MinConceptScore
        {
            get { return m_MinConceptScore; }
            set { if (SetSalienceOption(2001, value) == 0) m_MinConceptScore = value; }
        }

        private int m_ConceptWindowSize;
        public int ConceptWindowSize
        {
            get { return m_ConceptWindowSize; }
            set { if (SetSalienceOption(2002, value) == 0) m_ConceptWindowSize = value; }
        }

        private int m_ConceptTopicJump;
        public int ConceptTopicJump
        {
            get { return m_ConceptTopicJump; }
            set { if (SetSalienceOption(2003, value) == 0) m_ConceptTopicJump = value; }
        }

        private bool m_NongrammaticalTopics;
        public bool NongrammaticalTopics
        {
            get { return m_NongrammaticalTopics; }
            set { if (SetSalienceOption(2004, value) == 0) m_NongrammaticalTopics = value; }
        }

        private String m_ConceptTopics;
        /// <summary>
        /// Sets path to file containing concept topic definitions.
        /// </summary>
        public String ConceptTopics
        {
            get { return m_ConceptTopics; }
            set { if (SetSalienceOption(2005, value) == 0) m_ConceptTopics = value; }
        }
        #endregion // Concept options

        #region Entity options
        private bool m_AnaphoraResolution;
        /// <summary></summary>
        public bool AnaphoraResolution
        {
            get { return m_AnaphoraResolution; }
            set { if (SetSalienceOption(3001, value) == 0) m_AnaphoraResolution = value; }
        }

        private float m_ModelSensitivity;
        /// <summary></summary>
        public float ModelSensitivity
        {
            get { return m_ModelSensitivity; }
            set { if (SetSalienceOption(3002, value) == 0) m_ModelSensitivity = value; }
        }

        private int m_EntityThreshold;
        /// <summary>Integer property indicating the required threshold (between 0 and 100) that an entity must meet in order to be reported in entity results.</summary>
        /// <remarks>This property replaces the nThreshold member of the SE4 EntityParams structure.</remarks>
        public int EntityThreshold
        {
            get { return m_EntityThreshold; }
            set { if (SetSalienceOption(3003, value) == 0) m_EntityThreshold = value; }
        }

        private int m_EntitySummaryLength;
        /// <summary>Integer property indicating maximum length in sentences for entity summaries.</summary>
        /// <remarks>This property replaces both the bCalcSummary and nSummaryLength members of the SE4 EntityParams structure.</remarks>
        public int EntitySummaryLength
        {
            get { return m_EntitySummaryLength; }
            set { if (SetSalienceOption(3004, value) == 0) m_EntitySummaryLength = value; }
        }

        private bool m_EntityOverlap;
        /// <summary></summary>
        public bool EntityOverlap
        {
            get { return m_EntityOverlap; }
            set { if (SetSalienceOption(3005, value) == 0) m_EntityOverlap = value; }
        }

        private string m_EntityList;
        /// <summary>String property specifying a file containing user-defined entities.</summary>
        /// <remarks>This property replaces the SE4 method SetSimpleEntitiesFile.</remarks>
        public string EntityList
        {
            get { return m_EntityList; }
            set { if (SetSalienceOption(3006, value) == 0) m_EntityList = value; }
        }

        private bool m_SentimentThemeOverlap;
        /// <summary>
        /// Allows the sentiment phrases from an underlying LSF and custom HSD to overlap. This will allow sentiment phrases such as "good" and "good day" to both appear in results.
        /// </summary>
        public bool SentimentThemeOverlap
        {
            get { return m_SentimentThemeOverlap; }
            set { if (SetSalienceOption(3007, value) == 0) m_SentimentThemeOverlap = value; }
        }
        #endregion // Entity options

        #region Sentiment options
        #endregion // Sentiment options

        #region Topic options
        private bool m_TopicStemming;
        /// <summary>Boolean property indicating whether stemming should be applied to topic queries</summary>
        /// <remarks>This property replaces the TagStemming property introduced in SE4.</remarks>
        public bool TopicStemming
        {
            get { return m_TopicStemming; }
            set { if (SetSalienceOption(5000, value) == 0) m_TopicStemming = value; }
        }

        private string m_TopicList;
        /// <summary>String property specifying a file containing topic queries.</summary>
        /// <remarks>This property replaces the SE4 method SetTagDefinitions.</remarks>
        public string TopicList
        {
            get { return m_TopicList; }
            set { if (SetSalienceOption(5001, value) == 0) m_TopicList = value; }        
        }
        #endregion // Topic options

#region MANAGED .NET LAYER
    #region Constructors

        /// <summary>
        /// Standard constructor, requires path to license file and path to data directory
        /// </summary>
        /// <param name="sLicensePath">Path to a valid Lexalytics license file</param>
        /// <param name="sDataPath">Path to data directory to use for this session</param>
        /// <exception cref="SalienceException">
        /// Thrown if license file cannot be loaded or error in loading data directory (AccessViolationException), or unable to load underlying Salience Engine (DllNotFoundException)
        /// </exception>
        /// <see cref="Insert URL for wrapper documentation"/>
        /// <seealso cref="Insert URL for lxaLoadLicense and lxaOpenSalienceSession documentation"/>
        public SalienceFiveEngine(string sLicensePath, string sDataPath)
        {
            SalienceStartup oStartup = new SalienceStartup() { sDataDirectory = sDataPath, sUserDirectory = sDataPath + "/user", sLogPath = String.Empty };
            m_pSessionHandle = IntPtr.Zero;
            m_pLicense = IntPtr.Zero;
            int nRet;
            try
            {
                nRet = lxaLoadLicense(sLicensePath, ref m_pLicense);
                openSalienceSession(oStartup);
            }
            catch (AccessViolationException e)
            {
                throw new SalienceException("Could not load license at " + sLicensePath, e);
            }
            catch (DllNotFoundException e)
            {
                throw new SalienceException("Unable to load SalienceEngineFive.dll. This file must be in the same directory as the application.", e);
            }
            SetInternalCallback();
            //Check nError.
        }

        /// <summary>
        /// Standard constructor, requires path to license file and path to data directory
        /// </summary>
        /// <param name="sLicensePath">Path to a valid Lexalytics license file</param>
        /// <param name="sDataPath">Path to data directory to use for this session</param>
        /// <param name="sUserDir">Path to a user directory of customizations to use for this session</param>
        /// <exception cref="SalienceException">
        /// Thrown if license file cannot be loaded or error in loading data directory (AccessViolationException), or unable to load underlying Salience Engine (DllNotFoundException)
        /// </exception>
        /// <see cref="Insert URL for wrapper documentation"/>
        /// <seealso cref="Insert URL for lxaLoadLicense and lxaOpenSalienceSession documentation"/>
        public SalienceFiveEngine(string sLicensePath, string sDataPath, string sUserDir)
        {
            SalienceStartup oStartup = new SalienceStartup() { sDataDirectory = sDataPath, sUserDirectory = sUserDir, sLogPath = String.Empty };
            m_pSessionHandle = IntPtr.Zero;
            m_pLicense = IntPtr.Zero;
            int nRet;
            try
            {
                nRet = lxaLoadLicense(sLicensePath, ref m_pLicense);
                openSalienceSession(oStartup);
            }
            catch (AccessViolationException e)
            {
                throw new SalienceException("Could not load license at " + sLicensePath, e);
            }
            catch (DllNotFoundException e)
            {
                throw new SalienceException("Unable to load SalienceEngineFive.dll. This file must be in the same directory as the application.", e);
            }
            SetInternalCallback();
            //Check nError.
        }

        /// <summary>
        /// Standard constructor, requires path to license file and path to data directory
        /// </summary>
        /// <param name="sLicensePath">Path to a valid Lexalytics license file</param>
        /// <param name="sDataPath">Path to data directory to use for this session</param>
        /// <param name="sUserDir">Path to a user directory of customizations to use for this session</param>
        /// <param name="sLogPath">Path to a log file to write session in itialization information to</param>
        /// <exception cref="SalienceException">
        /// Thrown if license file cannot be loaded or error in loading data directory (AccessViolationException), or unable to load underlying Salience Engine (DllNotFoundException)
        /// </exception>
        /// <see cref="Insert URL for wrapper documentation"/>
        /// <seealso cref="Insert URL for lxaLoadLicense and lxaOpenSalienceSession documentation"/>
        public SalienceFiveEngine(string sLicensePath, string sDataPath, string sUserDir, string sLogPath)
        {
            SalienceStartup oStartup = new SalienceStartup() { sDataDirectory = sDataPath, sUserDirectory = sUserDir, sLogPath = sLogPath };
            m_pSessionHandle = IntPtr.Zero;
            m_pLicense = IntPtr.Zero;
            int nRet;
            try
            {
                nRet = lxaLoadLicense(sLicensePath, ref m_pLicense);
                openSalienceSession(oStartup);
            }
            catch (AccessViolationException e)
            {
                throw new SalienceException("Could not load license at " + sLicensePath, e);
            }
            catch (DllNotFoundException e)
            {
                throw new SalienceException("Unable to load SalienceEngineFive.dll. This file must be in the same directory as the application.", e);
            }
            SetInternalCallback();
            //Check nError.
        }

        /// <summary>
        /// Standard constructor, requires path to license file and path to data directory
        /// </summary>
        /// <param name="sLicensePath">Path to a valid Lexalytics license file</param>
        /// <param name="oStartup">Startup object giving other session startup parameters</param>
        /// <exception cref="SalienceException">
        /// Thrown if license file cannot be loaded or error in loading data directory (AccessViolationException), or unable to load underlying Salience Engine (DllNotFoundException)
        /// </exception>
        /// <see cref="Insert URL for wrapper documentation"/>
        /// <seealso cref="Insert URL for lxaLoadLicense and lxaOpenSalienceSession documentation"/>
        public SalienceFiveEngine(string sLicensePath, SalienceStartup oStartup)
        {
            m_pSessionHandle = IntPtr.Zero;
            m_pLicense = IntPtr.Zero;
            int nRet;
            try
            {
                nRet = lxaLoadLicense(sLicensePath, ref m_pLicense);
                openSalienceSession(oStartup);
            }
            catch (AccessViolationException e)
            {
                throw new SalienceException("Could not load license at " + sLicensePath, e);
            }
            catch (DllNotFoundException e)
            {
                throw new SalienceException("Unable to load SalienceEngineFive.dll. This file must be in the same directory as the application.", e);
            }
            SetInternalCallback();
            //Check nError.
        }

    #endregion

    #region Destructor
        ~SalienceFiveEngine()
        {
            Dispose(false);
        }
        protected void Dispose(bool disposing)
        {
            if (disposing)
            {
                // Code to dispose the managed resources of the class
            }
            // Code to dispose the un-managed resources of the class
            if (this.m_pSessionHandle != IntPtr.Zero)
            {
                lxaCloseSalienceSession(this.m_pSessionHandle);
            }
            if (this.m_pLicense != IntPtr.Zero)
            {
                lxaFreeLicense(ref m_pLicense);
            }

            m_bDisposed = true;
        }
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }
    #endregion

    #region Utility methods
        protected static string GetDefaultSalienceLocation()
        {
            return getDefaultSalienceLocation();
        }

        /// <summary>
        /// Wrapper method to return the Salience Engine version
        /// </summary>
        /// <returns>String with Salience Engine version information</returns>
        /// <remarks>Wrapper around C API method lxaGetSalienceVersion</remarks>
        /// <exception cref="SalienceException">Thrown if version information is not available</exception>
        /// <see cref="insert URL for .Net wrapper documentation"/>
        /// <seealso cref="insert URL for C API documentation"/>
        protected string GetSalienceVersion()
        {
            return getVersion();
        }

        /// <summary>
        /// Wrapper method to return information about the Salience Engine environment
        /// </summary>
        /// <returns>String with Salience Engine environment information</returns>
        /// <remarks>Wrapper around C API method lxaDumpEnvironment</remarks>
        /// <see cref="insert URL for .Net wrapper documentation"/>
        /// <seealso cref="insert URL for C API documentation"/>
        protected string DumpEnvironment()
        {
            return dumpEnvironment();
        }

        /// <summary>
        /// Returns an integer error code indicating last warning encountered by Salience Engine, generally in the process of text preparation
        /// </summary>
        /// <returns>Integer return code</returns>
        /// <remarks>Wrapper around C API method lxaGetLastWarnings</remarks>
        /// <see cref="insert URL for .Net wrapper documentation"/>
        /// <seealso cref="insert URL for C API documentation"/>
        protected int GetLastWarnings()
        {
            int nRet;
            lxaGetLastWarnings(m_pSessionHandle, out nRet);
            return nRet;
        }

        /// <summary>
        /// Adds a hand-scored dictionary to the engine for use in sentiment analysis.
        /// </summary>
        /// <param name="sDictionary">Path to a hand-scored dictionary, or string contents of a hand-scored dictionary.</param>
        /// <param name="bReset">Set to true to reset dictionaries and only use this one, set to false to add dictionary to others loaded.</param>
        /// <returns>Integer return code</returns>
        /// <remarks>Equivalent to SE4 method lxaAddSentimentDictionary, which is now an option that takes an additional parameter.</remarks>
        /// <see cref="insert URL for .Net wrapper documentation"/>
        /// <seealso cref="insert URL for C API documentation"/>
        protected int AddSentimentDictionary(string sDictionary, bool bReset)
        {
            int nValue = 0;
            if (!bReset)
                nValue = 1;

            return SetSalienceOption(4000, sDictionary, nValue);
        }

        /// <summary>
        /// Sets a Salience Engine option which takes an integer value
        /// </summary>
        /// <param name="nOption">Option constant as defined in SalienceOptions.h</param>
        /// <param name="nValue">Applicable integer value for the option</param>
        /// <exception cref="SalienceException">Thrown if invalid option specified or invalid value</exception>
        /// <returns>0 on success, otherwise throws SalienceException</returns>
        /// <remarks>Wrapper around C API method lxaSetSalienceOption</remarks>
        /// <see cref="insert URL to wrapper documentation"/>
        /// <seealso cref="insert URL to C API documetation"/>
        protected int SetSalienceOption(int nOption, int nValue)
        {
            int nRet = setSalienceOption(nOption, nValue);
            if (nRet == 0)
            {
                return nRet;
            }
            else
            {
                string errMessage = String.Format("Error setting option #{0}", nOption);
                switch (nRet)
                {
                    case 4:
                        errMessage += "Invalid parameter provided when setting Salience option.";
                        break;
                    case 12:
                        errMessage += "Option not supported in Salience options.";
                        break;
                    default:
                        break;
                }
                SalienceException errOption = new SalienceException(errMessage);
                throw errOption;
            }
        }

        /// <summary>
        /// Sets a Salience Engine option which takes an boolean value
        /// </summary>
        /// <param name="nOption">Option constant as defined in SalienceOptions.h</param>
        /// <param name="nValue">Applicable boolean value for the option</param>
        /// <exception cref="SalienceException">Thrown if invalid option specified or invalid value</exception>
        /// <returns>0 on success, otherwise throws SalienceException</returns>
        /// <remarks>Wrapper around C API method lxaSetSalienceOption</remarks>
        /// <see cref="insert URL to wrapper documentation"/>
        /// <seealso cref="insert URL to C API documetation"/>
        protected int SetSalienceOption(int nOption, bool bValue)
        {
            int nRet = setSalienceOption(nOption, bValue);
            if (nRet == 0)
            {
                return nRet;
            }
            else
            {
                string errMessage = String.Format("Error setting option #{0}", nOption);
                switch (nRet)
                {
                    case 4:
                        errMessage += "Invalid parameter provided when setting Salience option.";
                        break;
                    case 12:
                        errMessage += "Option not supported in Salience options.";
                        break;
                    default:
                        break;
                }
                SalienceException errOption = new SalienceException(errMessage);
                throw errOption;
            }
        }

        /// <summary>
        /// Sets a Salience Engine option which takes a float value
        /// </summary>
        /// <param name="nOption">Option constant as defined in SalienceOptions.h</param>
        /// <param name="nValue">Applicable float value for the option</param>
        /// <exception cref="SalienceException">Thrown if invalid option specified or invalid value</exception>
        /// <returns>0 on success, otherwise throws SalienceException</returns>
        /// <remarks>Wrapper around C API method lxaSetSalienceOption</remarks>
        /// <see cref="insert URL to wrapper documentation"/>
        /// <seealso cref="insert URL to C API documetation"/>
        protected int SetSalienceOption(int nOption, float fValue)
        {
            int nRet = setSalienceOption(nOption, fValue);
            if (nRet == 0)
            {
                return nRet;
            }
            else
            {
                string errMessage = String.Format("Error setting option #{0}", nOption);
                switch (nRet)
                {
                    case 4:
                        errMessage += "Invalid parameter provided when setting Salience option.";
                        break;
                    case 12:
                        errMessage += "Option not supported in Salience options.";
                        break;
                    default:
                        break;
                }
                SalienceException errOption = new SalienceException(errMessage);
                throw errOption;
            }
        }

        /// <summary>
        /// Sets a Salience Engine option which takes a character string value
        /// </summary>
        /// <param name="nOption">Option constant as defined in SalienceOptions.h</param>
        /// <param name="sValue">String value for the option</param>
        /// <exception cref="SalienceException">Thrown if invalid option specified or invalid value</exception>
        /// <returns>0 on success, otherwise throws SalienceException</returns>
        /// <remarks>Wrapper around C API method lxaSetSalienceOption</remarks>
        /// <see cref="insert URL to wrapper documentation"/>
        /// <seealso cref="insert URL to C API documetation"/>
        protected int SetSalienceOption(int nOption, string sValue)
        {
            int nRet = setSalienceOption(nOption, sValue);
            if (nRet == 0)
            {
                return nRet;
            }
            else
            {
                string errMessage = String.Format("Error setting option #{0}", nOption);
                switch (nRet)
                {
                    case 4:
                        errMessage += "Invalid parameter provided when setting Salience option.";
                        break;
                    case 12:
                        errMessage += "Option not supported in Salience options.";
                        break;
                    default:
                        break;
                }
                SalienceException errOption = new SalienceException(errMessage);
                throw errOption;
            }
        }

        /// <summary>
        /// Sets a Salience Engine option which takes a character string value and integer value
        /// </summary>
        /// <param name="nOption">Option constant as defined in SalienceOptions.h</param>
        /// <param name="sValue">String value for the option</param>
        /// <param name="nValue">Integer value for the option</param>
        /// <exception cref="SalienceException">Thrown if invalid option specified or invalid value</exception>
        /// <returns>0 on success, otherwise throws SalienceException</returns>
        /// <remarks>Wrapper around C API method lxaSetSalienceOption</remarks>
        /// <see cref="insert URL to wrapper documentation"/>
        /// <seealso cref="insert URL to C API documetation"/>
        protected int SetSalienceOption(int nOption, string sValue, int nValue)
        {
            int nRet = setSalienceOption(nOption, sValue, nValue);
            if (nRet == 0)
            {
                return nRet;
            }
            else
            {
                string errMessage = String.Format("Error setting option #{0}", nOption);
                switch (nRet)
                {
                    case 4:
                        errMessage += "Invalid parameter provided when setting Salience option.";
                        break;
                    case 12:
                        errMessage += "Option not supported in Salience options.";
                        break;
                    default:
                        break;
                }
                SalienceException errOption = new SalienceException(errMessage);
                throw errOption;
            }
        }


    #endregion // Utility methods

    #region Text preparation methods

        /// <summary>
        /// Wrapper method to submit text to Salience Engine for processing
        /// </summary>
        /// <param name="sText">The text to be processed</param>
        /// <returns>Integer error return code</returns>
        /// <remarks>Wrapper around C API method lxaPrepareText</remarks>
        /// <exception cref="SalienceException">Thrown if text cannot be parsed</exception>
        /// <see cref="insert URL for .Net wrapper documentation"/>
        /// <seealso cref="insert URL for C API documentation"/>
        protected int PrepareText(string sText)
        {
            int nRet = processPrepareText(sText);
            if (nRet != 0 && nRet != 6)
            {
                throw new SalienceException(getLastError());
            }
            return nRet;
        }

        /// <summary>
        /// Wrapper method to submit contents of a file to Salience Engine for processing
        /// </summary>
        /// <param name="sFile">Full path to a readable text file</param>
        /// <returns>Integer error return code</returns>
        /// <remarks>Wrapper around C API method lxaPrepareTextFromFile</remarks>
        /// <exception cref="SalienceException">Thrown if file cannot be found or text cannot be parsed</exception>
        /// <see cref="insert URL for .Net wrapper documentation"/>
        /// <seealso cref="insert URL for C API documentation"/>
        protected int PrepareTextFromFile(string sFile)
        {
            int nRet = processPrepareTextFromFile(sFile);
            if (nRet != 0 && nRet != 6)
            {
                throw new SalienceException(getLastError());
            }
            return nRet;
        }

        /// <summary>
        /// Creates a collection from a list of strings
        /// </summary>
        /// <returns>Integer error return code</returns>
        /// <remarks>Wrapper around C API method lxaPrepareCollection</remarks>
        /// <exception cref="SalienceException">Thrown if text cannot be parsed</exception>
        /// <see cref="insert URL for .Net wrapper documentation"/>
        /// <seealso cref="insert URL for C API documentation"/>
        protected int PrepareCollectionFromList(String Name, List<String> lstText)
        {
            List<SalienceCollectionDocument> CollectionDocuments = new List<SalienceCollectionDocument>();
            for (int i = 0; i < lstText.Count; i++)
            {
                CollectionDocuments.Add(new SalienceCollectionDocument() { sIdentifier = i.ToString(), sText = lstText[i] });
            }
            int nRet = processPrepareCollectionFromList(Name, CollectionDocuments);
            if (nRet != 0 && nRet != 6)
            {
                throw new SalienceException(getLastError());
            }
            return nRet;
        }

        /// <summary>
        /// Creates a collection from a file containing content
        /// </summary>
        /// <param name="sName">Name for the collection</param>
        /// <param name="sFile">Full path to a readable text file</param>
        /// <returns>Integer error return code</returns>
        /// <remarks>Wrapper around C API method lxaPrepareCollectionFromFile</remarks>
        /// <exception cref="SalienceException">Thrown if text cannot be parsed</exception>
        /// <see cref="insert URL for .Net wrapper documentation"/>
        /// <seealso cref="insert URL for C API documentation"/>
        protected int PrepareCollectionFromFile(string sName, string sFile)
        {
            int nRet = processPrepareCollectionFromFile(sName, sFile);
            if (nRet != 0 && nRet != 6)
            {
                throw new SalienceException(getLastError());
            }
            return nRet;
        }

    #endregion // Text preparation methods

    #region Details methods
        /// <summary>
        /// Wrapper method to return details about current document
        /// </summary>
        /// <returns>SalienceDocumentDetails object</returns>
        /// <remarks>Wrapper around C API method lxaGetDocumentDetails</remarks>
        /// <see cref="insert URL for .Net wrapper documentation"/>
        /// <seealso cref="insert URL for C API documentation"/>
        public SalienceDocumentDetails GetDocumentDetails()
        {
            return processDocumentDetails();
        }

        /// <summary>
        /// Wrapper method to return details about the current collection
        /// </summary>
        /// <returns>SalienceCollectionDetails object</returns>
        /// <remarks>Wrapper around C API method lxaGetCollectionDetails</remarks>
        /// <see cref="insert URL for .Net wrapper documentation"/>
        /// <seealso cref="insert URL for C API documentation"/>
        public SalienceCollectionDetails GetCollectionDetails()
        {
            return processCollectionDetails();
        }

    #endregion // Details methods

    #region Document methods
        /// <summary>
        /// Returns a summary for the document
        /// </summary>
        /// <param name="nLength">The length in sentences that the summary should be.</param>
        /// <returns>A text string containing a summary of the document up to nLength sentences.</returns>
        /// <remarks>Wrapper around C API method lxaGetSummary</remarks>
        /// <see cref="insert URL for .Net wrapper documentation"/>
        /// <seealso cref="insert URL for C API documentation"/>
        protected string GetSummary(int nLength)
        {
            return processSummary(nLength);
        }

        /// <summary>
        /// Returns the sentiment analysis for the document
        /// </summary>
        /// <param name="bUseChains">A boolean indicating whether or not to use lexical chains for sentiment phrase analysis</param>
        /// <returns>A SalienceSentiment object containing members with sentiment results</returns>
        /// <remarks>Wrapper around C API method lxaGetSentiment</remarks>
        /// <see cref="insert URL for .Net wrapper documentation"/>
        /// <seealso cref="insert URL for C API documentation"/>
        protected SalienceSentiment GetDocumentSentiment(bool bUseChains)
        {
            return processDocumentSentiment(bUseChains);
        }

        /// <summary>
        /// Returns a list of themes extracted from the document
        /// </summary>
        /// <returns>A List of SalienceTheme objects, each containing information about a theme extracted from the document</returns>
        /// <remarks>Wrapper around C API method lxaGetThemes</remarks>
        /// <see cref="insert URL for .Net wrapper documentation"/>
        /// <seealso cref="insert URL for C API documentation"/>
        protected List<SalienceTheme> GetDocumentThemes()
        {
            return processDocumentThemes();
        }

        /// <summary>
        /// Returns a list of topics extracted from the document via tag queries
        /// </summary>
        /// <returns>A List of SalienceTopic objects, each containing information about a topic extracted from the document through the tag queries specified for the session</returns>
        /// <remarks>Wrapper around C API method lxaGetQueryDefinedTopics</remarks>
        /// <see cref="insert URL for .Net wrapper documentation"/>
        /// <seealso cref="insert URL for C API documentation"/>
        protected List<SalienceTopic> GetQueryDefinedTopics()
        {
            return processDocumentTopics();
        }

        /// <summary>
        /// Returns a list of topics extracted from the document via the concept matrix
        /// </summary>
        /// <returns>A List of SalienceTopic objects, each containing information about a topic identified for the document through the use of the concept matrix</returns>
        /// <remarks>Wrapper around C API method lxaGetConceptDefinedTopics</remarks>
        /// <see cref="insert URL for .Net wrapper documentation"/>
        /// <seealso cref="insert URL for C API documentation"/>
        protected List<SalienceTopic> GetConceptDefinedTopics()
        {
            return processDocumentConceptTopics();
        }

        /// <summary>
        /// Returns a string listing the top contributing words and phrases from the current document for each Concept Topic Match.
        /// </summary>
        /// <returns>A text string with concept topic results, and contributing words/phrases.</returns>
        /// <remarks>Wrapper around C API method lxaExplainConceptMatches</remarks>
        /// <see cref="insert URL for .Net wrapper documentation"/>
        /// <seealso cref="insert URL for C API documentation"/>
        protected string ExplainConceptMatches()
        {
            return processExplainConceptMatches();
        }

    #endregion // Document methods

    #region Entity methods
        /// <summary>
        /// Returns a list of named entities extracted from the document.
        /// This is equivalent to the SE4.4 method GetEntities.
        /// </summary>
        /// <param name="Params">A structure with entity extraction options</param>
        /// <returns>A List of SalienceEntity objects, each containing information about a named entity</returns>
        /// <remarks>Wrapper around C API method lxaGetNamedEntities</remarks>
        /// <see cref="insert URL for .Net wrapper documentation"/>
        /// <seealso cref="insert URL for C API documentation"/>
        protected List<SalienceEntity> GetNamedEntities(SalienceEntityParams Params)
        {
            EntityThreshold = Params.nRequiredThreshold;
            EntitySummaryLength = Params.nSummaryLength;
            return GetNamedEntities();
        }

        /// <summary>
        /// Returns a list of named entities extracted from the document.
        /// This is equivalent to the SE4.4 method GetEntities.
        /// </summary>
        /// <returns>A List of SalienceEntity objects, each containing information about a named entity</returns>
        /// <remarks>Wrapper around C API method lxaGetNamedEntities</remarks>
        /// <see cref="insert URL for .Net wrapper documentation"/>
        /// <seealso cref="insert URL for C API documentation"/>
        protected List<SalienceEntity> GetNamedEntities()
        {
            return processNamedEntities();
        }

        /// <summary>
        /// Returns a list of user-defined entities extracted from the document.
        /// This is equivalent to the SE4.4 method GetSimpleEntities and GetQueryDefinedEntities.
        /// </summary>
        /// <returns>A List of SalienceEntity objects, each containing information about a user-defined entity</returns>
        /// <remarks>Wrapper around C API method lxaGetUserDefinedEntities</remarks>
        /// <see cref="insert URL for .Net wrapper documentation"/>
        /// <seealso cref="insert URL for C API documentation"/>
        protected List<SalienceEntity> GetUserDefinedEntities()
        {
            return processUserDefinedEntities();
        }

        /// <summary>
        /// Returns a list of relationships for named entities extracted from the document.
        /// This is equivalent to the SE4.4 method GetEntityRelationships.
        /// </summary>
        /// <returns>A List of SalienceRelationship objects, each containing information about a relationship between multiple named entities</returns>
        /// <remarks>Wrapper around C API method lxaGetNamedEntityRelationships</remarks>
        /// <see cref="insert URL for .Net wrapper documentation"/>
        /// <seealso cref="insert URL for C API documentation"/>
        protected List<SalienceRelationship> GetNamedEntityRelationships(int nThreshold)
        {
            EntityThreshold = nThreshold;
            return processRelationships();
        }

        /// <summary>
        /// Returns a list of opinions for named entities extracted from the document.
        /// This is equivalent to the SE4.4 method GetEntityOpinions.
        /// </summary>
        /// <returns>A List of SalienceOpinion objects, each containing information about an opinion related to a named entity</returns>
        /// <remarks>Wrapper around C API method lxaGetNamedEntityOpinions</remarks>
        /// <see cref="insert URL for .Net wrapper documentation"/>
        /// <seealso cref="insert URL for C API documentation"/>
        protected List<SalienceOpinion> GetNamedEntityOpinions(int nThreshold)
        {
            EntityThreshold = nThreshold;
            return processOpinions();
        }

        /// <summary>
        /// Returns a list of relationships for user-defined entities extracted from the document.
        /// This is equivalent to the SE4.4 method GetEntityRelationships for named entities, but for simple and query-defined entities.
        /// </summary>
        /// <returns>A List of SalienceRelationship objects, each containing information about a relationship between multiple user-defined entities</returns>
        /// <remarks>Wrapper around C API method lxaGetUserEntityRelationships</remarks>
        /// <see cref="insert URL for .Net wrapper documentation"/>
        /// <seealso cref="insert URL for C API documentation"/>
        protected List<SalienceRelationship> GetUserEntityRelationships()
        {
            return processUserEntityRelationships();
        }

        /// <summary>
        /// Returns a list of opinions for user-defined entities extracted from the document.
        /// This is equivalent to the SE4.4 method GetEntityOpinions for named entities, but for simple and query-defined entities.
        /// </summary>
        /// <returns>A List of SalienceOpinion objects, each containing information about an opinion related to a user-defined entity</returns>
        /// <remarks>Wrapper around C API method lxaGetUserEntityOpinions</remarks>
        /// <see cref="insert URL for .Net wrapper documentation"/>
        /// <seealso cref="insert URL for C API documentation"/>
        protected List<SalienceOpinion> GetUserEntityOpinions()
        {
            return processUserEntityOpinions();
        }

    #endregion 

    #region Collection methods
        /// <summary>
        /// Returns a list of themes extracted from the collection
        /// </summary>
        /// <returns>A List of SalienceTheme objects, each containing information about a theme extracted from the collection</returns>
        /// <remarks>Wrapper around C API method lxaGetCollectionThemes</remarks>
        /// <see cref="insert URL for .Net wrapper documentation"/>
        /// <seealso cref="insert URL for C API documentation"/>
        protected List<SalienceTheme> GetCollectionThemes()
        {
            return processCollectionThemes();
        }

        /// <summary>
        /// Returns a list of entities extracted from the collection
        /// </summary>
        /// <returns>A List of SalienceEntity objects, each containing information about an entity extracted from the collection</returns>
        /// <remarks>Wrapper around C API method lxaGetCollectionThemes</remarks>
        /// <see cref="insert URL for .Net wrapper documentation"/>
        /// <seealso cref="insert URL for C API documentation"/>
        protected List<SalienceCollectionEntity> GetCollectionEntities()
        {
            return processCollectionEntities();
        }

        /// <summary>
        /// Returns a list of facets extracted from the collection
        /// </summary>
        /// <returns>A List of SalienceFacet objects, each containing information about a facet and its attributes extracted from the collection</returns>
        /// <remarks>Wrapper around C API method lxaGetCollectionFacets</remarks>
        /// <see cref="insert URL for .Net wrapper documentation"/>
        /// <seealso cref="insert URL for C API documentation"/>
        protected List<SalienceFacet> GetCollectionFacets()
        {
            return processCollectionFacets();
        }

        /// <summary>
        /// Returns a list of topics extracted from the collection via tag query definitions.
        /// </summary>
        /// <returns>A List of SalienceTopic objects, each containing information about a topic identified for at least one document in the collection</returns>
        /// <remarks>Wrapper around C API method lxaGetCollectionQueryDefinedTopics</remarks>
        /// <see cref="insert URL for .Net wrapper documentation"/>
        /// <seealso cref="insert URL for C API documentation"/>
        protected List<SalienceTopic> GetCollectionQueryDefinedTopics()
        {
            return processCollectionQueryTopics();
        }

        /// <summary>
        /// Returns a list of topics extracted from the collection via the concept matrix.
        /// </summary>
        /// <returns>A List of SalienceTopic objects, each containing information about a topic identified for at least one document in the collection</returns>
        /// <remarks>Wrapper around C API method lxaGetCollectionConceptDefinedTopics</remarks>
        /// <see cref="insert URL for .Net wrapper documentation"/>
        /// <seealso cref="insert URL for C API documentation"/>
        protected List<SalienceTopic> GetCollectionConceptTopics()
        {
            return processCollectionConceptTopics();
        }

    #endregion // Collection methods

    #region Markup methods
        protected string GetEntityTaggedText()
        {
            return processNamedEntityTaggedText();
        }

        protected string GetUserEntityTaggedText()
        {
            return processUserEntityTaggedText();
        }

        protected string GetPOSTaggedText()
        {
            return processPOSTaggedText();
        }

        protected string GetSentimentTaggedText()
        {
            return processSentimentTaggedText();
        }

        protected string GetUserOpinionTaggedText()
        {
            return processUserOpinionTaggedText();
        }

        protected string GetNamedOpinionTaggedText()
        {
            return processNamedOpinionTaggedText();
        }

    #endregion // Markup methods

    #region Reinitialize methods

        public void ReinitializeThemes()
        {
            SetSalienceOption(1008, 1);
        }

    #endregion

#endregion MANAGED .NET LAYER

        #region UNMANAGED .NET TO C LAYER (INTERNAL IMPLEMENTATION)

        #region Session methods
        private unsafe void openSalienceSession(SalienceStartup oStartup)
        {
            umSalienceStartup umStartup = new umSalienceStartup();
            umStartup.acDataDirectory = oStartup.sDataDirectory;
            umStartup.acUserDirectory = oStartup.sUserDirectory;
            
            if (oStartup.sLogPath != String.Empty)
            {
                umStartup.nStartupLog = 1;
                umStartup.acLogPath = Marshal.StringToHGlobalAnsi(oStartup.sLogPath);
            }
            
            int nError = lxaOpenSalienceSession(m_pLicense, ref umStartup, ref m_pSessionHandle);
            if (nError != 0)
            {
               throw new SalienceException(umStartup.acError);
            }
        }

    #endregion // Session methods

    #region Text preparation methods
        private unsafe int processPrepareText(string sText)
        {
            return lxaPrepareText(m_pSessionHandle, sText);
        }

        private unsafe int processPrepareTextFromFile(string sFile)
        {
            return lxaPrepareTextFromFile(m_pSessionHandle, sFile);
        }

        private unsafe int processPrepareCollectionFromList(string sCollectionName, List<SalienceCollectionDocument> lstDocs)
        {
            Callback_Current = Callback_PrepareCollection;
            umSalienceCollection Collection = new umSalienceCollection();
            Collection.nSize = lstDocs.Count;
            IntPtr pArray = Marshal.AllocHGlobal(lstDocs.Count * Marshal.SizeOf(typeof(umSalienceCollectionDocument)));
            Collection.pDocuments = (umSalienceCollectionDocument*)pArray.ToPointer();
            Collection.acName = Marshal.StringToHGlobalAnsi(sCollectionName);
            for (int i = 0; i < Collection.nSize; i++)
            {
                Collection.pDocuments[i].acIndentifier = Marshal.StringToHGlobalAnsi(lstDocs[i].sIdentifier);
                Collection.pDocuments[i].acText = Marshal.StringToHGlobalAnsi(lstDocs[i].sText);
                Collection.pDocuments[i].nIsText = 1;
                Collection.pDocuments[i].nSplitByLine = 0;
            }
            int nRet = lxaPrepareCollection(m_pSessionHandle, &Collection);
            for (int i = 0; i < Collection.nSize; i++)
            {
                Marshal.FreeHGlobal(Collection.pDocuments[i].acIndentifier);
                Marshal.FreeHGlobal(Collection.pDocuments[i].acText);
            }
            Marshal.FreeHGlobal(Collection.acName);
            Marshal.FreeHGlobal(pArray);
            return nRet;
        }

        private unsafe int processPrepareCollectionFromFile(string sCollectionName, string sFile)
        {
            return lxaPrepareCollectionFromFile(m_pSessionHandle, sCollectionName, sFile);
        }
    #endregion // Text preparation methods

    #region Details methods
        private unsafe SalienceDocumentDetails processDocumentDetails()
        {
            umSalienceDocumentDetails DocDetails = new umSalienceDocumentDetails();
            lxaGetDocumentDetails(m_pSessionHandle, &DocDetails);
            SalienceDocumentDetails ReturnDetails = new SalienceDocumentDetails();
            ReturnDetails.sText = MarshalPtrToUtf8(DocDetails.acInternalRepresentation);
            ReturnDetails.sFingerprint = MarshalPtrToUtf8(DocDetails.acFingerprint);
            ReturnDetails.nSentenceCount = DocDetails.nSentenceCount;
            ReturnDetails.nWordCount = DocDetails.nWordCount;
            ReturnDetails.TermFrequency = processTokens(DocDetails.oTermFrequency);
            ReturnDetails.BiGramFrequency = processTokens(DocDetails.oBiGrams);
            ReturnDetails.TriGramFrequency = processTokens(DocDetails.oTriGrams);
            ReturnDetails.Chunks = processChunks(DocDetails.pSentences, DocDetails.nSentenceCount);
            lxaFreeDocumentDetails(&DocDetails);

            return ReturnDetails;
        }

        private unsafe SalienceCollectionDetails processCollectionDetails()
        {
            umSalienceCollectionDetails CollectionDetails = new umSalienceCollectionDetails();
            lxaGetCollectionDetails(m_pSessionHandle, &CollectionDetails);
            SalienceCollectionDetails ReturnDetails = new SalienceCollectionDetails();
            ReturnDetails.nSize = CollectionDetails.nSize;
            return ReturnDetails;
        }

        private unsafe List<SalienceToken> processTokens(umSalienceTokenList TokenList)
        {
            List<SalienceToken> ReturnTokens = new List<SalienceToken>(TokenList.nLength);
            umSalienceToken* pToken = TokenList.pTokens;
            for (int i = 0; i < TokenList.nLength; i++, pToken++)
            {
                SalienceToken Token = new SalienceToken();
                Token.nCount = pToken->nCount;
                Token.sToken = MarshalPtrToUtf8(pToken->acToken);
                ReturnTokens.Add(Token);
            }
            return ReturnTokens;
        }

        private unsafe List<SalienceChunk> processChunks(umSalienceSentence* pSentences, int nLength)
        {
            List<SalienceChunk> ReturnChunks = new List<SalienceChunk>(nLength);
            for (int i = 0; i < nLength; i++, pSentences++)
            {
                for (int j = 0; j < pSentences->nChunkCount; j++)
                {
                    umSalienceChunk oChunk = pSentences->pChunks[j];
                    SalienceChunk Chunk = new SalienceChunk();
                    Chunk.nSentence = i;
                    Chunk.sLabel = MarshalPtrToUtf8(oChunk.acLabel);

                    Chunk.Tokens = new List<SalienceWord>(oChunk.nLength);
                    umSalienceWord* pToken = oChunk.pTokens;
                    for (int k = 0; k < oChunk.nLength; k++, pToken++)
                    {
                        SalienceWord Token = new SalienceWord();
                        Token.sPOSTag = MarshalPtrToUtf8(pToken->acPOSTag);
                        Token.sToken = MarshalPtrToUtf8(pToken->acToken);
                        Token.bInvert = true;
                        if (pToken->nInvert == 0)
                        {
                            Token.bInvert = false;
                        }


                        Chunk.Tokens.Add(Token);
                    }

                    ReturnChunks.Add(Chunk);
                }
            }
            return ReturnChunks;
        }

    #endregion // Details methods

    #region Document methods
        private unsafe string processSummary(int nLength)
        {
            IntPtr acSummary;
            lxaGetSummary(m_pSessionHandle, nLength, out acSummary);
            if (acSummary == IntPtr.Zero)
            {
                throw new SalienceException("Could not extract summary from current document.");
            }
            string text = MarshalPtrToUtf8(acSummary);
            lxaFreeString(acSummary);
            return text;
        }

        private unsafe SalienceSentiment processDocumentSentiment(bool bUseChains)
        {
            umSalienceSentimentResult oResult = new umSalienceSentimentResult();
            lxaGetSentiment(m_pSessionHandle, bUseChains ? 1 : 0, &oResult);

            SalienceSentiment oRet = new SalienceSentiment();
            oRet.fScore = oResult.fScore;
            oRet.Phrases = new List<SalienceSentimentPhrase>(oResult.oPhrases.nLength);
            umSalienceSentimentPhrase* pPhrase = oResult.oPhrases.pPhrases;
            for (int i = 0; i < oResult.oPhrases.nLength; i++, pPhrase++)
            {
                SalienceSentimentPhrase Phrase = new SalienceSentimentPhrase();
                Phrase.fScore = pPhrase->fScore;
                Phrase.Phrase = processPhrase(pPhrase->oPhrase);
                Phrase.nType = pPhrase->nType;
                Phrase.sSource = Marshal.PtrToStringAnsi(pPhrase->acSource);
                Phrase.nNegated = pPhrase->nNegated;
                Phrase.SupportingPhrase = processPhrase(pPhrase->oSupportingPhrase);
                oRet.Phrases.Add(Phrase);
            }

            //Now the other bits
            for (int i = 0; i < oResult.nModelCount; i++)
            {
                SalienceModelSentiment Sentiment = new SalienceModelSentiment();
                Sentiment.fPositive = oResult.pModel[i].fPositive;
                Sentiment.fNegative = oResult.pModel[i].fNegative;
                Sentiment.fNeutral = oResult.pModel[i].fNeutral;
                Sentiment.fMixed = oResult.pModel[i].fMixed;
                Sentiment.nBest = oResult.pModel[i].nBest;
                Sentiment.sName = Marshal.PtrToStringAnsi(oResult.pModel[i].acModelName);
                oRet.ModelSentiment.Add(Sentiment);

            }

            lxaFreeSentimentResult(&oResult);
            return oRet;
        }

        private unsafe List<SalienceTheme> processDocumentThemes()
        {
            umSalienceThemeList ThemeList = new umSalienceThemeList();
            lxaGetThemes(m_pSessionHandle, ref ThemeList);
            return processThemes(ThemeList, true);
        }

        private unsafe List<SalienceTopic> processDocumentTopics()
        {
            umSalienceTopicList oTopicList = new umSalienceTopicList();
            lxaGetQueryDefinedTopics(m_pSessionHandle, &oTopicList);

            List<SalienceTopic> vRet = new List<SalienceTopic>(oTopicList.nLength);
            umSalienceTopic* pTopic = oTopicList.pTopics;
            for (int i = 0; i < oTopicList.nLength; i++, pTopic++)
            {
                SalienceTopic Topic = new SalienceTopic();
                Topic.fScore = pTopic->fScore;
                Topic.nHits = pTopic->nHits;
                Topic.sTopic = MarshalPtrToUtf8(pTopic->acTopic);
                vRet.Add(Topic);
            }
            lxaFreeTopicList(&oTopicList);
            return vRet;

        }

        private unsafe List<SalienceTopic> processDocumentConceptTopics()
        {
            umSalienceTopicList oTopicList = new umSalienceTopicList();
            lxaGetConceptDefinedTopics(m_pSessionHandle, &oTopicList);

            List<SalienceTopic> vRet = new List<SalienceTopic>(oTopicList.nLength);
            umSalienceTopic* pTopic = oTopicList.pTopics;
            for (int i = 0; i < oTopicList.nLength; i++, pTopic++)
            {
                SalienceTopic Topic = new SalienceTopic();
                Topic.fScore = pTopic->fScore;
                Topic.nHits = pTopic->nHits;
                Topic.sTopic = MarshalPtrToUtf8(pTopic->acTopic);
                vRet.Add(Topic);
            }
            lxaFreeTopicList(&oTopicList);
            return vRet;
        }

        private unsafe string processExplainConceptMatches()
        {
            IntPtr acBuffer;
            lxaExplainConceptMatches(m_pSessionHandle, out acBuffer);
            if (acBuffer == IntPtr.Zero)
            {
                throw new SalienceException("Error attempting to run lxaExplainConceptMatches");
            }
            string text = MarshalPtrToUtf8(acBuffer);
            lxaFreeString(acBuffer);
            return text;
        }

    #endregion // Document methods

    #region Entity methods

        private unsafe List<SalienceEntity> processNamedEntities()
        {
            umSalienceEntityList EntityList = new umSalienceEntityList();
            lxaGetNamedEntities(m_pSessionHandle, &EntityList);
            return processEntityList(EntityList, true);
        }

        private unsafe List<SalienceEntity> processUserDefinedEntities()
        {
            umSalienceEntityList EntityList = new umSalienceEntityList();
            lxaGetUserDefinedEntities(m_pSessionHandle, &EntityList);
            return processEntityList(EntityList, true);
        }

        private unsafe List<SalienceRelationship> processRelationships()
        {
            umSalienceRelationList RelationList = new umSalienceRelationList();
            lxaGetNamedEntityRelationships(m_pSessionHandle, &RelationList);

            List<SalienceRelationship> vRet = new List<SalienceRelationship>(RelationList.nLength);
            umSalienceRelation* pRelation = RelationList.pRelations;
            for (int i = 0; i < RelationList.nLength; i++, pRelation++)
            {
                SalienceRelationship Relationship = new SalienceRelationship();
                Relationship.fScore = pRelation->fConfidence;
                Relationship.sLabel = MarshalPtrToUtf8(pRelation->acType);
                Relationship.sExtra = MarshalPtrToUtf8(pRelation->acExtra);
                Relationship.Entities = processEntityList(pRelation->oEntities, false);

                vRet.Add(Relationship);
            }
            lxaFreeRelationList(&RelationList);
            return vRet;
        }

        private unsafe List<SalienceOpinion> processOpinions()
        {
            umSalienceOpinionList OpinionList = new umSalienceOpinionList();
            lxaGetNamedEntityOpinions(m_pSessionHandle, &OpinionList);
            List<SalienceOpinion> vRet = new List<SalienceOpinion>(OpinionList.nLength);

            umSalienceOpinion* pOpinion = OpinionList.pOpinions;
            for (int i = 0; i < OpinionList.nLength; i++, pOpinion++)
            {
                SalienceOpinion Opinion = new SalienceOpinion();
                Opinion.sQuotation = MarshalPtrToUtf8(pOpinion->acQuotation);
                Opinion.fSentiment = pOpinion->fSentiment;
                Opinion.Speaker = processEntity(&pOpinion->oSpeaker);
                if (pOpinion->nHasTheme != 0)
                {
                    Opinion.bThemeOpinion = true;
                    Opinion.ThemeTopic = processTheme(&pOpinion->oThemeTopic);
                }
                else
                {
                    Opinion.bThemeOpinion = false;
                    Opinion.EntityTopic = processEntity(&pOpinion->oEntityTopic);
                }
                vRet.Add(Opinion);
            }
            return vRet;

        }

        private unsafe List<SalienceRelationship> processUserEntityRelationships()
        {
            umSalienceRelationList RelationList = new umSalienceRelationList();
            lxaGetUserEntityRelationships(m_pSessionHandle, &RelationList);

            List<SalienceRelationship> vRet = new List<SalienceRelationship>(RelationList.nLength);
            umSalienceRelation* pRelation = RelationList.pRelations;
            for (int i = 0; i < RelationList.nLength; i++, pRelation++)
            {
                SalienceRelationship Relationship = new SalienceRelationship();
                Relationship.fScore = pRelation->fConfidence;
                Relationship.sLabel = MarshalPtrToUtf8(pRelation->acType);
                Relationship.sExtra = MarshalPtrToUtf8(pRelation->acExtra);
                Relationship.Entities = processEntityList(pRelation->oEntities, false);

                vRet.Add(Relationship);
            }
            lxaFreeRelationList(&RelationList);
            return vRet;
        }

        private unsafe List<SalienceOpinion> processUserEntityOpinions()
        {
            umSalienceOpinionList OpinionList = new umSalienceOpinionList();
            lxaGetUserEntityOpinions(m_pSessionHandle, &OpinionList);
            List<SalienceOpinion> vRet = new List<SalienceOpinion>(OpinionList.nLength);

            umSalienceOpinion* pOpinion = OpinionList.pOpinions;
            for (int i = 0; i < OpinionList.nLength; i++, pOpinion++)
            {
                SalienceOpinion Opinion = new SalienceOpinion();
                Opinion.sQuotation = MarshalPtrToUtf8(pOpinion->acQuotation);
                Opinion.fSentiment = pOpinion->fSentiment;
                Opinion.Speaker = processEntity(&pOpinion->oSpeaker);
                if (pOpinion->nHasTheme != 0)
                {
                    Opinion.bThemeOpinion = true;
                    Opinion.ThemeTopic = processTheme(&pOpinion->oThemeTopic);
                }
                else
                {
                    Opinion.bThemeOpinion = false;
                    Opinion.EntityTopic = processEntity(&pOpinion->oEntityTopic);
                }
                vRet.Add(Opinion);
            }
            return vRet;

        }

        private unsafe SalienceEntity processEntity(umSalienceEntity* pEntity)
        {
            SalienceEntity Entity = new SalienceEntity();
            Entity.fSentimentScore = pEntity->fSentimentScore;
            Entity.nAbout = pEntity->nAbout;
            Entity.nConfident = pEntity->nConfident;
            Entity.nEvidence = pEntity->nEvidence;
            Entity.nCount = pEntity->oMentions.nLength;
            Entity.nFirstPos = pEntity->oMentions.pMentions[0].oPhrase.nWord;
            Entity.sLabel = MarshalPtrToUtf8(pEntity->acLabel);
            Entity.sNormalizedForm = MarshalPtrToUtf8(pEntity->acNormalizedForm);
            Entity.sSummary = MarshalPtrToUtf8(pEntity->acSummary);
            Entity.sType = MarshalPtrToUtf8(pEntity->acType);

            Entity.Mentions = processMentions(pEntity->oMentions);
            Entity.Themes = processThemes(pEntity->oThemes, false);
            Entity.SentimentPhrases = processSentimentPhrases(pEntity->oSentimentPhrases);
            return Entity;
        }

        private unsafe List<SalienceEntity> processEntityList(umSalienceEntityList EntityList, bool bFree)
        {
            List<SalienceEntity> ReturnEntities = new List<SalienceEntity>(EntityList.nLength);
            umSalienceEntity* pEntity = EntityList.pEntities;
            for (int i = 0; i < EntityList.nLength; i++, pEntity++)
            {
                SalienceEntity Entity = processEntity(pEntity);
                ReturnEntities.Add(Entity);
            }
            if (bFree)
            {
                lxaFreeEntityList(&EntityList);
            }
            return ReturnEntities;

        }

        private unsafe List<SalienceMention> processMentions(umSalienceMentionList MentionList)
        {
            List<SalienceMention> ReturnMentions = new List<SalienceMention>(MentionList.nLength);
            umSalienceMention* pMention = MentionList.pMentions;
            for (int i = 0; i < MentionList.nLength; i++, pMention++)
            {
                SalienceMention Mention = new SalienceMention();
                Mention.fScore = pMention->fScore;
                Mention.nType = pMention->nType;
                Mention.Phrase = processPhrase(pMention->oPhrase);
                ReturnMentions.Add(Mention);
            }
            return ReturnMentions;
        }

        private unsafe List<SaliencePhrase> processPhraseMentions(umSaliencePhraseList MentionList)
        {
            List<SaliencePhrase> ReturnPhrases = new List<SaliencePhrase>(MentionList.nLength);
            umSaliencePhrase* pPhraseMention = MentionList.pPhrases;
            for (int i = 0; i < MentionList.nLength; i++, pPhraseMention++)
            {
                SaliencePhrase PhraseMention = new SaliencePhrase();
                PhraseMention = processPhrase(MentionList.pPhrases[i]);
                ReturnPhrases.Add(PhraseMention);
            }
            return ReturnPhrases;
        }

    #endregion // Entity methods

    #region Shared Entity and Document methods
        private unsafe List<SalienceSentimentPhrase> processSentimentPhrases(umSalienceSentimentPhraseList oPhraseList)
        {
            List<SalienceSentimentPhrase> oRet = new List<SalienceSentimentPhrase>(oPhraseList.nLength);
            umSalienceSentimentPhrase* pPhrase = oPhraseList.pPhrases;
            for (int i = 0; i < oPhraseList.nLength; i++, pPhrase++)
            {
                SalienceSentimentPhrase Phrase = new SalienceSentimentPhrase();
                Phrase.fScore = pPhrase->fScore;
                Phrase.Phrase = processPhrase(pPhrase->oPhrase);
                Phrase.nType = pPhrase->nType;
                Phrase.nNegated = pPhrase->nNegated;
                Phrase.SupportingPhrase = processPhrase(pPhrase->oSupportingPhrase);
                oRet.Add(Phrase);
            }

            return oRet;

        }

        private unsafe List<SalienceTheme> processThemes(umSalienceThemeList ThemeList, bool bFree)
        {
            List<SalienceTheme> ReturnThemes = new List<SalienceTheme>(ThemeList.nLength);
            umSalienceTheme* pTheme = ThemeList.pThemes;
            for (int i = 0; i < ThemeList.nLength; i++, pTheme++)
            {

                ReturnThemes.Add(processTheme(pTheme));
            }
            //GetEntities calls ProcessThemes to fill its Theme List, but also calls FreeEntity when it's done, which calls lxaFreeThemes.
            //Thus we call processThemes(true) only when we have a themes list that's not connected to an entity.
            if (bFree)
            {
                lxaFreeThemeList(&ThemeList);
            }
            return ReturnThemes;
        }

        private unsafe SalienceTheme processTheme(umSalienceTheme* pTheme)
        {
            SalienceTheme Theme = new SalienceTheme();
            Theme.fScore = pTheme->fScore;
            Theme.fSentiment = pTheme->fSentiment;
            Theme.nEvidence = pTheme->nEvidence;
            Theme.nThemeType = pTheme->nThemeType;
            Theme.nEvidence = pTheme->nEvidence;
            Theme.bAbout = (pTheme->nAbout == 1);
            Theme.sStemmedTheme = MarshalPtrToUtf8(pTheme->acStemmedTheme);
            Theme.sTheme = MarshalPtrToUtf8(pTheme->acTheme);
            Theme.sNormalizedTheme = MarshalPtrToUtf8(pTheme->acNormalizedTheme);
            Theme.sSummary = MarshalPtrToUtf8(pTheme->acSummary);
            Theme.ThemeMentions = processPhraseMentions(pTheme->oMentions);
            return Theme;
        }

        private unsafe SaliencePhrase processPhrase(umSaliencePhrase umPhrase)
        {
            SaliencePhrase Phrase = new SaliencePhrase();
            Phrase.nByte = umPhrase.nByte;
            Phrase.nByteLength = umPhrase.nByteLength;
            Phrase.nLength = umPhrase.nLength;
            Phrase.nDocument = umPhrase.nDocument;
            Phrase.nSentence = umPhrase.nSentence;
            Phrase.nWord = umPhrase.nWord;
            Phrase.sText = MarshalPtrToUtf8(umPhrase.acText);
            return Phrase;
        }

    #endregion // Shared Entity and Document methods

    #region Collection methods
        private unsafe List<SalienceTheme> processCollectionThemes()
        {
            Callback_Current = Callback_GetCollectionThemes;
            umSalienceThemeList ThemeList = new umSalienceThemeList();
            lxaGetCollectionThemes(m_pSessionHandle, ref ThemeList);
            return processThemes(ThemeList, true);
        }

        private unsafe List<SalienceCollectionEntity> processCollectionEntities()
        {
            Callback_Current = Callback_GetCollectionEntities;
            umSalienceCollectionEntityList EntityList = new umSalienceCollectionEntityList();
            lxaGetCollectionEntities(m_pSessionHandle, &EntityList);
            return processCollectionEntityList(EntityList, true);
        }

        private unsafe List<SalienceCollectionEntity> processCollectionEntityList(umSalienceCollectionEntityList EntityList, bool bFree)
        {
            Callback_Current = Callback_GetCollectionEntities;
            List<SalienceCollectionEntity> ReturnEntities = new List<SalienceCollectionEntity>(EntityList.nLength);
            umSalienceCollectionEntity* pEntity = EntityList.pCollectionEntities;
            for (int i = 0; i < EntityList.nLength; i++, pEntity++)
            {
                SalienceCollectionEntity Entity = processCollectionEntity(pEntity);
                ReturnEntities.Add(Entity);
            }
            if (bFree)
            {
                lxaFreeCollectionEntityList(&EntityList);
            }
            return ReturnEntities;
        }

        private unsafe SalienceCollectionEntity processCollectionEntity(umSalienceCollectionEntity* pEntity)
        {
            SalienceCollectionEntity Entity = new SalienceCollectionEntity();
            Entity.nCount = pEntity->nCount;
            Entity.nNegativeCount = pEntity->nNegativeCount;
            Entity.nNeutralCount = pEntity->nNeutralCount;
            Entity.nPositiveCount = pEntity->nPositiveCount;
            Entity.sLabel = MarshalPtrToUtf8(pEntity->acLabel);
            Entity.sNormalizedForm = MarshalPtrToUtf8(pEntity->acNormalizedForm);
            Entity.sType = MarshalPtrToUtf8(pEntity->acType);
            Entity.oMentions = processPhraseMentions(pEntity->oMentions);
            return Entity;
        }

        private unsafe List<SalienceFacet> processCollectionFacets()
        {
            Callback_Current = Callback_GetCollectionFacets;
            umSalienceFacetList FacetList = new umSalienceFacetList();
            lxaGetCollectionFacets(m_pSessionHandle, &FacetList);
            return processCollectionFacetList(FacetList, true);
        }

        private unsafe List<SalienceFacet> processCollectionFacetList(umSalienceFacetList FacetList, bool bFree)
        {
            Callback_Current = Callback_GetCollectionFacets;
            List<SalienceFacet> ReturnFacets = new List<SalienceFacet>(FacetList.nLength);
            umSalienceFacet* pFacet = FacetList.pFacetList;
            for (int i = 0; i < FacetList.nLength; i++, pFacet++)
            {
                SalienceFacet Facet = processFacet(pFacet);
                ReturnFacets.Add(Facet);
            }
            if (bFree)
            {
                lxaFreeFacetList(&FacetList);
            }
            return ReturnFacets;
        }

        private unsafe SalienceFacet processFacet(umSalienceFacet* pFacet)
        {
            SalienceFacet Facet = new SalienceFacet();
            Facet.nCount = pFacet->nCount;
            Facet.nNegativeCount = pFacet->nNegativeCount;
            Facet.nNeutralCount = pFacet->nNeutralCount;
            Facet.nPositiveCount = pFacet->nPositiveCount;
            Facet.sFacet = MarshalPtrToUtf8(pFacet->acFacet);
            Facet.Mentions = processPhraseMentions(pFacet->oMentions);
            Facet.Attributes = new List<SalienceAttribute>();
            for (int i = 0; i < pFacet->nSubjectLength; i++ )
            {
                SalienceAttribute Attribute = new SalienceAttribute();
                Attribute.sAttribute = MarshalPtrToUtf8(pFacet->pAttributes[i].acAttribute);
                Attribute.nCount = pFacet->pAttributes[i].nCount;
                Attribute.Mentions = processPhraseMentions(pFacet->pAttributes[i].oMentions);
                Facet.Attributes.Add(Attribute);
            }
            return Facet;
        }

        private unsafe List<SalienceTopic> processCollectionQueryTopics()
        {
            umSalienceTopicList oTopicList = new umSalienceTopicList();
            lxaGetCollectionQueryDefinedTopics(m_pSessionHandle, &oTopicList);

            List<SalienceTopic> vRet = new List<SalienceTopic>(oTopicList.nLength);
            umSalienceTopic* pTopic = oTopicList.pTopics;
            for (int i = 0; i < oTopicList.nLength; i++, pTopic++)
            {
                SalienceTopic Topic = new SalienceTopic();
                Topic.fScore = pTopic->fScore;
                Topic.nHits = pTopic->nHits;
                Topic.sTopic = MarshalPtrToUtf8(pTopic->acTopic);
                string sMentions = MarshalPtrToUtf8(pTopic->acAdditional);
                Topic.vDocuments = new List<string>(sMentions.Split('|'));
                vRet.Add(Topic);
            }
            lxaFreeTopicList(&oTopicList);
            return vRet;
        }

        private unsafe List<SalienceTopic> processCollectionConceptTopics()
        {
            umSalienceTopicList oTopicList = new umSalienceTopicList();
            lxaGetCollectionConceptDefinedTopics(m_pSessionHandle, &oTopicList);

            List<SalienceTopic> vRet = new List<SalienceTopic>(oTopicList.nLength);
            umSalienceTopic* pTopic = oTopicList.pTopics;
            for (int i = 0; i < oTopicList.nLength; i++, pTopic++)
            {
                SalienceTopic Topic = new SalienceTopic();
                Topic.fScore = pTopic->fScore;
                Topic.nHits = pTopic->nHits;
                Topic.sTopic = MarshalPtrToUtf8(pTopic->acTopic);
                string sMentions = MarshalPtrToUtf8(pTopic->acAdditional);
                Topic.vDocuments = new List<string>(sMentions.Split('|'));
                vRet.Add(Topic);
            }
            lxaFreeTopicList(&oTopicList);
            return vRet;
        }

    #endregion // Collection methods

    #region Utility methods
        private static unsafe string getDefaultSalienceLocation()
        {
            IntPtr acPath;
            lxaGetDefaultSalienceLocation(out acPath);
            if (acPath == IntPtr.Zero) throw new SalienceException("Could not find default Salience location.");
            string sPath = Marshal.PtrToStringAnsi(acPath);
            lxaFreeString(acPath);
            return sPath;
        }

        private unsafe string getVersion()
        {
            IntPtr acVersion;
            lxaGetSalienceVersion(out acVersion);
            if (acVersion == IntPtr.Zero) throw new SalienceException("Could not read version information.");
            string sVersion = Marshal.PtrToStringAnsi(acVersion);
            lxaFreeString(acVersion);
            return sVersion;
        }

        private unsafe string getLastError()
        {
            IntPtr acError;
            lxaGetSalienceErrorString(m_pSessionHandle, out acError);
            string sError = Marshal.PtrToStringAnsi(acError);
            lxaFreeString(acError);
            return sError;
        }

        private unsafe string dumpEnvironment()
        {
            IntPtr acEnvironment;
            lxaDumpEnvironment(m_pSessionHandle, out acEnvironment);
            string sEnvironment = Marshal.PtrToStringAnsi(acEnvironment);
            lxaFreeString(acEnvironment);
            return sEnvironment;
        }

        private unsafe int setSalienceOption(int nOption, int nValue)
        {
            umSalienceOption umOption = new umSalienceOption();
            umOption.nOption = nOption;   
            umOption.nValue = nValue;

            return lxaSetSalienceOption(m_pSessionHandle, &umOption);
        }

        private unsafe int setSalienceOption(int nOption, bool bValue)
        {
            umSalienceOption umOption = new umSalienceOption();
            umOption.nOption = nOption;   
            if (bValue)
                umOption.nValue = 0;
            else
                umOption.nValue = 1;

            return lxaSetSalienceOption(m_pSessionHandle, &umOption);
        }

        private unsafe int setSalienceOption(int nOption, float fValue)
        {
            umSalienceOption umOption = new umSalienceOption();
            umOption.nOption = nOption;
            umOption.fValue = fValue;

            return lxaSetSalienceOption(m_pSessionHandle, &umOption);
        }

        private unsafe int setSalienceOption(int nOption, string sValue)
        {
            umSalienceOption umOption = new umSalienceOption();
            umOption.nOption = nOption;
            umOption.acValue = Marshal.StringToHGlobalAnsi(sValue);

            return lxaSetSalienceOption(m_pSessionHandle, &umOption);
        }

        private unsafe int setSalienceOption(int nOption, string sValue, int nValue)
        {
            umSalienceOption umOption = new umSalienceOption();
            umOption.nOption = nOption;
            umOption.nValue = nValue;
            umOption.acValue = Marshal.StringToHGlobalAnsi(sValue);

            return lxaSetSalienceOption(m_pSessionHandle, &umOption);
        }
        protected dEngineNotification Callback_Current;

        protected unsafe void SetInternalCallback()
        {
            fCallback = new umSalienceCallback(handleCallbacks);
            IntPtr pCallback = Marshal.GetFunctionPointerForDelegate(fCallback);
            lxaSetSalienceCallback(m_pSessionHandle,pCallback,null);
        }

        protected unsafe int handleCallbacks(void* pClass, int nId, IntPtr acMessage)
        {
            if (Callback_Current != null)
            {
                string sMessage = MarshalPtrToUtf8(acMessage);
                Callback_Current.Invoke(nId, sMessage);
            }
            return 0;
        }

    #endregion // Utility methods

    #region Marked Text methods

        private unsafe string entityTagHelper(umSalienceDocument oDoc)
        {
            StringBuilder sMarkup = new StringBuilder();
	        int nPrimaryId = -1;
	        int nSecondaryId = -1;
	        string sType = "";
	        for(int i = 0; i < oDoc.nSentenceCount; i++)
	        {
		        for(int j = 0; j < oDoc.pSentences[i].nLength; j++)
		        {
			        if(oDoc.pSentences[i].pTokens[j].nId == -1)
			        {
				        if(nPrimaryId != -1)
				        {
                            sMarkup.Append("[/");
					        sMarkup.Append(sType);
					        sMarkup.Append("]");
					        nPrimaryId = -1;
					        nSecondaryId = -1;
				        }
                        if (oDoc.pSentences[i].pTokens[j].nPostFixed == 0)
                        {
                            sMarkup.Append(" ");
                        }
				        sMarkup.Append(MarshalPtrToUtf8(oDoc.pSentences[i].pTokens[j].acToken));
			        }
			        else
			        {
				        if(nPrimaryId == oDoc.pSentences[i].pTokens[j].nId && nSecondaryId == oDoc.pSentences[i].pTokens[j].nSecondaryId)
				        {
                            if (oDoc.pSentences[i].pTokens[j].nPostFixed == 0)
                            {
                                sMarkup.Append(" ");
                            }
					        sMarkup.Append(MarshalPtrToUtf8(oDoc.pSentences[i].pTokens[j].acToken));
				        }
				        else
				        {
					        if(nPrimaryId != -1)
					        {
						        sMarkup.Append("[/");
						        sMarkup.Append(sType);
						        sMarkup.Append("]");
					        }
                            if (oDoc.pSentences[i].pTokens[j].nPostFixed == 0)
                            {
                                sMarkup.Append(" ");
                            }
					        nPrimaryId = oDoc.pSentences[i].pTokens[j].nId;
					        nSecondaryId = oDoc.pSentences[i].pTokens[j].nSecondaryId;
					        sType = MarshalPtrToUtf8(oDoc.pSentences[i].pTokens[j].acEntityType);
                            sMarkup.Append("[");
					        sMarkup.Append(sType);
					        sMarkup.Append("]");
					        sMarkup.Append(MarshalPtrToUtf8(oDoc.pSentences[i].pTokens[j].acToken));
				        }
			        }
		        }
	        }
	        if(nPrimaryId != -1)
	        {
                sMarkup.Append("[/");
		        sMarkup.Append(sType);
		        sMarkup.Append("]");
		        nPrimaryId = -1;
		        nSecondaryId = -1;
	        }
	        return sMarkup.ToString();
        }

        private unsafe string processNamedEntityTaggedText()
        {
            umSalienceDocument oDoc;
            lxaGetNamedEntityMarkup(m_pSessionHandle,&oDoc);
            return entityTagHelper(oDoc);
        }

        private unsafe string processUserEntityTaggedText()
        {
            umSalienceDocument oDoc;
            lxaGetUserEntityMarkup(m_pSessionHandle, &oDoc);
            return entityTagHelper(oDoc);
        }

        private unsafe string processPOSTaggedText()
        {
            StringBuilder sReturn = new StringBuilder();
            umSalienceDocument oDoc;
            lxaGetPOSMarkup(m_pSessionHandle, &oDoc);

            for (int i = 0; i < oDoc.nSentenceCount; i++)
            {
                for (int j = 0; j < oDoc.pSentences[i].nLength; j++)
                {
                    if (oDoc.pSentences[i].pTokens[j].nPostFixed == 0)
                    {
                        sReturn.Append(" ");
                    }
                    string sTag = MarshalPtrToUtf8(oDoc.pSentences[i].pTokens[j].acPOSTag);
                    sReturn.Append("[" + sTag + "]");
                    sReturn.Append(MarshalPtrToUtf8(oDoc.pSentences[i].pTokens[j].acToken));
                    sReturn.Append("[/" + sTag + "]");
                }
            }
            return sReturn.ToString();
        }

        private unsafe string processSentimentTaggedText(float fNegative=-0.3f, float fPositive = 0.3f)
        {
            StringBuilder sReturn = new StringBuilder();
            umSalienceDocument oDoc;
            lxaGetSentimentMarkup(m_pSessionHandle, &oDoc);
            string sLastTag = "";
            for (int i = 0; i < oDoc.nSentenceCount; i++)
            {
                if (oDoc.pSentences[i].nPolar == 1)
                {
                    sReturn.Append("[LXA_SENTENCE_POLAR]");
                }
                
                
                for (int j = 0; j < oDoc.pSentences[i].nLength; j++)
                {
                    string sTag = "";
                    int nType = oDoc.pSentences[i].pTokens[j].nSentimentType;
                    if(nType != -1)
                    {
                        if(nType == 0)
                        {
                            sTag = "LXA_STOP";
                        }
                        else if(nType == 2)
                        {
                            sTag = "LXA_POSSIBLE";
                        }
                        else if(oDoc.pSentences[i].pTokens[j].nInvert != 0)
                        {
                            sTag = "LXA_INVERT";
                        }
                        else
                        {
                            if(nType == 1)
                            {
                                sTag = "LXA_HANDSCORE_";
                            }
                            else
                            {
                                sTag = "LXA_INTERNAL_";
                            }
                            if(oDoc.pSentences[i].pTokens[j].fSentiment < fNegative)
                            {
                                sTag += "NEGATIVE";
                            }
                            else if (oDoc.pSentences[i].pTokens[j].fSentiment > fPositive)
                            {
                                sTag += "POSITIVE";
                            }
                            else
                            {
                                sTag += "NEUTRAL";
                            }
                        }
                    }
                    if (sTag != sLastTag)
                    {
                        if (sLastTag != "")
                        {
                            sReturn.Append("[/" + sLastTag + "]");
                        }
                        if (oDoc.pSentences[i].pTokens[j].nPostFixed == 0)
                        {
                            sReturn.Append(" ");
                        }
                        if (sTag != "")
                        {
                            sReturn.Append("[" + sTag + "]");
                        }
                        sLastTag = sTag;
                        sReturn.Append(MarshalPtrToUtf8(oDoc.pSentences[i].pTokens[j].acToken));
                    }
                    else
                    {
                        if (oDoc.pSentences[i].pTokens[j].nPostFixed == 0)
                        {
                            sReturn.Append(" ");
                        }
                        sReturn.Append(MarshalPtrToUtf8(oDoc.pSentences[i].pTokens[j].acToken));
                    }
                }
                if (sLastTag != "")
                {
                    sReturn.Append("[/" + sLastTag + "]");
                }

                if (oDoc.pSentences[i].nPolar == 1)
                {
                    sReturn.Append("[/LXA_SENTENCE_POLAR]");
                }
                
            }
            return sReturn.ToString();
        }

      private unsafe string processNamedOpinionTaggedText()
      {
            IntPtr acTaggedText;
            lxaGetNamedOpinionTaggedText(m_pSessionHandle, out acTaggedText);
            string sTaggedText = MarshalPtrToUtf8(acTaggedText);
            lxaFreeString(acTaggedText);
            return sTaggedText;
       }


      private unsafe string processUserOpinionTaggedText()
      {
          IntPtr acTaggedText;
          lxaGetUserOpinionTaggedText(m_pSessionHandle, out acTaggedText);
          string sTaggedText = MarshalPtrToUtf8(acTaggedText);
          lxaFreeString(acTaggedText);
          return sTaggedText;
      }


    #endregion // Marked Text methods

    #region Marshaling methods
        /// <summary>
        /// gets around .NET's lack of native utf8 support. given a pointer to a utf8 char* string, turns it into a byte[]
        /// </summary>
        /// <param name="pText">an IntPtr representing a char*</param>
        /// <returns>the bytes of the char*</returns>
        unsafe static byte[] Utf8byteArray(IntPtr pText)
        {
            if (pText == IntPtr.Zero) return new byte[0];

            //Get pText's pointer, cast it to a byte.
            byte* ptr = (byte*)pText.ToPointer();
            int length = 0;
            //read until we find a null character in the text.
            for (length = 0; ptr[length] != 0; length++) ;

            //now that we've got the size, copy from the pointer to an array.
            byte[] utf8 = new byte[length];
            for (int i = 0; i < utf8.Length; i++)
            {
                utf8[i] = ptr[i];
            }
            //converted! System.Text.Encoding.UTF8.GetString can now be used to get text
            //We don't return that string, because the byte array ends up being useful later in GetDocument.
            return utf8;
        }

        unsafe static string MarshalPtrToUtf8(IntPtr pText)
        {
            //We can marshal to ASCII and unicode, but not UTF8. For ease of use I've written this marshal function.
            return System.Text.Encoding.UTF8.GetString(Utf8byteArray(pText));
        }

    #endregion 

#endregion // UNMANAGED .NET TO C LAYER (INTERNAL IMPLEMENTATION)

#region UNMANAGED C LAYER (CONNECTION TO UNDERLYING SALIENCE C API)
        private const string dllName = "SalienceFive.dll";

    #region Global methods
        [DllImport(dllName)]
        private unsafe static extern int lxaLoadLicense(string acPath, ref IntPtr pLicense);

        [DllImport(dllName)]
        private unsafe static extern int lxaFreeLicense(ref IntPtr pLicense);

        [DllImport(dllName)]
        private unsafe static extern int lxaFreeString(IntPtr acBuffer);

    #endregion // Global methods

    #region Session methods
        [DllImport(dllName)]
        private unsafe static extern int lxaOpenSalienceSession(IntPtr pLicense, ref umSalienceStartup pStartup, ref IntPtr ppSession);

        [DllImport(dllName)]
        private unsafe static extern int lxaCloseSalienceSession(IntPtr pSession);

    #endregion // Session methods

    #region Prepare functions
        [DllImport(dllName)]
        private unsafe static extern int lxaPrepareText(IntPtr pSession, string acText);

        [DllImport(dllName)]
        private unsafe static extern int lxaPrepareTextFromFile(IntPtr pSession, string acFile);

        // TODO: SE5 Documentation, new API method
        [DllImport(dllName)]
        private unsafe static extern int lxaPrepareCollection(IntPtr pSession, umSalienceCollection* pCollection);

        // TODO: SE5 Documentation, new API method
        [DllImport(dllName)]
        private unsafe static extern int lxaPrepareCollectionFromFile(IntPtr pSession, string acName, string acFile);

    #endregion // Prepare functions

    #region Details functions
        [DllImport(dllName)]
        private unsafe static extern int lxaGetDocumentDetails(IntPtr pSession, umSalienceDocumentDetails* pDetails);

        // TODO: SE5 Documentation, new API method
        [DllImport(dllName)]
        private unsafe static extern int lxaGetCollectionDetails(IntPtr pSession, umSalienceCollectionDetails* pDetails);

    #endregion // Details functions

    #region Document level functions
        [DllImport(dllName)]
        private unsafe static extern int lxaGetSummary(IntPtr pSession, int nLength, out IntPtr ppBuffer);

        [DllImport(dllName)]
        private unsafe static extern int lxaGetSentiment(IntPtr pSession, int nUseChains, umSalienceSentimentResult* pResult);

        [DllImport(dllName)]
        private unsafe static extern int lxaGetThemes(IntPtr pSession, ref umSalienceThemeList pResults);

        [DllImport(dllName)]
        private unsafe static extern int lxaGetNamedEntities(IntPtr pSession, umSalienceEntityList* pResults);

        [DllImport(dllName)]
        private unsafe static extern int lxaGetUserDefinedEntities(IntPtr pSession, umSalienceEntityList* pResults);

        [DllImport(dllName)]
        private unsafe static extern int lxaGetQueryDefinedTopics(IntPtr pSession, umSalienceTopicList* pTopicList);

        [DllImport(dllName)]
        private unsafe static extern int lxaGetConceptDefinedTopics(IntPtr pSession, umSalienceTopicList* pTopicList);

        [DllImport(dllName)]
        private unsafe static extern int lxaExplainConceptMatches(IntPtr pSession, out IntPtr ppBuffer);

        [DllImport(dllName)]
        private unsafe static extern int lxaGetNamedEntityRelationships(IntPtr pSession, umSalienceRelationList* pResults);

        [DllImport(dllName)]
        private unsafe static extern int lxaGetNamedEntityOpinions(IntPtr pSession, umSalienceOpinionList* pResults);

        [DllImport(dllName)]
        private unsafe static extern int lxaGetUserEntityRelationships(IntPtr pSession, umSalienceRelationList* pResults);

        [DllImport(dllName)]
        private unsafe static extern int lxaGetUserEntityOpinions(IntPtr pSession, umSalienceOpinionList* pResults);

    #endregion // Document level functions

    #region Collection level functions
        [DllImport(dllName)]
        private unsafe static extern int lxaGetCollectionThemes(IntPtr pSession, ref umSalienceThemeList pResults);

        [DllImport(dllName)]
        private unsafe static extern int lxaGetCollectionFacets(IntPtr pSession, umSalienceFacetList* pResults);

        [DllImport(dllName)]
        private unsafe static extern int lxaGetCollectionQueryDefinedTopics(IntPtr pSession, umSalienceTopicList* pResults);

        [DllImport(dllName)]
        private unsafe static extern int lxaGetCollectionConceptDefinedTopics(IntPtr pSession, umSalienceTopicList* pResults);

        [DllImport(dllName)]
        private unsafe static extern int lxaGetCollectionEntities(IntPtr pSession, umSalienceCollectionEntityList* pResults);

    #endregion // Collection level functions

    #region Markup functions
        [DllImport(dllName)]
        private unsafe static extern int lxaGetNamedEntityMarkup(IntPtr pSession, umSalienceDocument* pDocument);

        [DllImport(dllName)]
        private unsafe static extern int lxaGetUserEntityMarkup(IntPtr pSession, umSalienceDocument* pDocument);

        [DllImport(dllName)]
        private unsafe static extern int lxaGetPOSMarkup(IntPtr pSession, umSalienceDocument* pDocument);

        [DllImport(dllName)]
        private unsafe static extern int lxaGetSentimentMarkup(IntPtr pSession, umSalienceDocument* pDocument);

        [DllImport(dllName)]
        private unsafe static extern int lxaGetNamedOpinionTaggedText(IntPtr pSession, out IntPtr acTaggedText);

        [DllImport(dllName)]
        private unsafe static extern int lxaGetUserOpinionTaggedText(IntPtr pSession, out IntPtr acTaggedText);

    #endregion // Markup functions

    #region Free functions
        [DllImport(dllName)]
        private unsafe static extern int lxaFreeEntityList(umSalienceEntityList* pResults);

        [DllImport(dllName)]
        private unsafe static extern int lxaFreeCollectionEntityList(umSalienceCollectionEntityList* pResults);

        [DllImport(dllName)]
        private unsafe static extern int lxaFreeThemeList(umSalienceThemeList* pThemesList);

        [DllImport(dllName)]
        private unsafe static extern int lxaFreeFacetList(umSalienceFacetList* pFacetList);

        [DllImport(dllName)]
        private unsafe static extern int lxaFreeRelationList(umSalienceRelationList* pRelationList);

        [DllImport(dllName)]
        private unsafe static extern int lxaFreeOpinionList(umSalienceOpinionList* pOpinionList);

        [DllImport(dllName)]
        private unsafe static extern int lxaFreeSentimentResult(umSalienceSentimentResult* pPhraseList);

        [DllImport(dllName)]
        private unsafe static extern int lxaFreeDocumentDetails(umSalienceDocumentDetails* pDetails);

        [DllImport(dllName)]
        private unsafe static extern int lxaFreeTopicList(umSalienceTopicList* pTopicList);

        [DllImport(dllName)]
        private unsafe static extern int lxaFreeDocument(umSalienceDocument* pDocument);

        [DllImport(dllName)]
        private unsafe static extern int lxaFreePhraseList(umSaliencePhraseList* pPhraseList);

    #endregion // Free methods

    #region Utility functions
        [DllImport(dllName)]
        private unsafe static extern int lxaGetDefaultSalienceLocation(out IntPtr ppBuffer);

        [DllImport(dllName)]
        private unsafe static extern int lxaGetSalienceVersion(out IntPtr ppBuffer);

        [DllImport(dllName)]
        private unsafe static extern int lxaGetSalienceErrorString(IntPtr pSession, out IntPtr ppErrorString);

        [DllImport(dllName)]
        private unsafe static extern int lxaGetLastWarnings(IntPtr pSession, out int nWarnings);

        [DllImport(dllName)]
        private unsafe static extern int lxaSetSalienceOption(IntPtr pSession, umSalienceOption* pOption);

        [DllImport(dllName)]
        private unsafe static extern int lxaSetSalienceCallback(IntPtr pSession, IntPtr pCallback, void* pParam);

        [DllImport(dllName)]
        private unsafe static extern int lxaDumpEnvironment(IntPtr pSession, out IntPtr acEnvironment);

    #endregion // Utility functions

#endregion // UNMANAGED C LAYER (CONNECTION TO UNDERLYING SALIENCE C API)


    }
}
